@@ -1,5 +1,6 @@
 /* 
  * Copyright (C) 2002 Red Hat, Inc.
+ * Copyright (C) 2009 Julien Danjou <julien@danjou.info>
  * 
  * Permission is hereby granted, free of charge, to any person
  * obtaining a copy of this software and associated documentation
@@ -22,13 +23,17 @@
  * SOFTWARE.
  */
 
+#include <config.h>
+
+#include <xcb/xcb_event.h>
+
 #include "sn-xmessages.h"
 #include "sn-list.h"
 #include "sn-internals.h"
 
 typedef struct
 {
-  Display       *xdisplay;
+  void          *xid;
   Window         root;
   Atom           type_atom;
   Atom           type_atom_begin;
@@ -63,8 +68,8 @@ sn_internal_add_xmessage_func (SnDisplay      *display,
   
   handler = sn_new0 (SnXmessageHandler, 1);
 
-  handler->xdisplay = sn_display_get_x_display (display);
-  handler->root = RootWindow (handler->xdisplay, screen);
+  handler->xid = sn_internal_display_get_id (display);
+  handler->root = sn_internal_display_get_root_window (display, screen);
   handler->type_atom = sn_internal_atom_get (display, message_type);
   handler->type_atom_begin = sn_internal_atom_get (display, message_type_begin);
   handler->message_type = sn_internal_strdup (message_type);
@@ -120,7 +125,7 @@ sn_internal_remove_xmessage_func (SnDisplay      *display,
   fhd.func = func;
   fhd.func_data = func_data;
   fhd.handler = NULL;
-  fhd.root = RootWindow (sn_display_get_x_display (display), screen);
+  fhd.root = sn_internal_display_get_root_window (display, screen);
   
   if (xmessage_funcs != NULL)
     sn_list_foreach (xmessage_funcs, find_handler_foreach, &fhd);
@@ -145,11 +150,6 @@ sn_internal_broadcast_xmessage   (SnDisplay      *display,
                                   const char     *message_type_begin,
                                   const char     *message)
 {
-  Atom type_atom;
-  Atom type_atom_begin;
-  Window xwindow;
-  Display *xdisplay;
-
   if (!sn_internal_utf8_validate (message, -1))
     {
       fprintf (stderr,
@@ -158,75 +158,138 @@ sn_internal_broadcast_xmessage   (SnDisplay      *display,
       return;
     }
 
-  xdisplay = sn_display_get_x_display (display);
-
+  switch (sn_internal_display_get_type (display))
   {
-    XSetWindowAttributes attrs;
-
-    attrs.override_redirect = True;
-    attrs.event_mask = PropertyChangeMask | StructureNotifyMask;
-
-    xwindow =
-      XCreateWindow (xdisplay,
-                     RootWindow (xdisplay, 0),
-                     -100, -100, 1, 1,
-                     0,
-                     CopyFromParent,
-                     CopyFromParent,
-                     CopyFromParent,
-                     CWOverrideRedirect | CWEventMask,
-                     &attrs);
-  }
-
-  type_atom = sn_internal_atom_get (display, message_type);
-  type_atom_begin = sn_internal_atom_get (display, message_type_begin);
+   case SN_DISPLAY_TYPE_XLIB:
+    {
+      Atom type_atom;
+      Atom type_atom_begin;
+      Window xwindow;
+      XSetWindowAttributes attrs;
+      Display *xdisplay = sn_display_get_x_display (display);
   
-  {
-    XEvent xevent;
-    const char *src;
-    const char *src_end;
-    char *dest;
-    char *dest_end;
-    
-    xevent.xclient.type = ClientMessage;
-    xevent.xclient.message_type = type_atom_begin;
-    xevent.xclient.display = xdisplay;
-    xevent.xclient.window = xwindow;
-    xevent.xclient.format = 8;
+      attrs.override_redirect = True;
+      attrs.event_mask = PropertyChangeMask | StructureNotifyMask;
 
-    src = message;
-    src_end = message + strlen (message) + 1; /* +1 to include nul byte */
+      xwindow =
+        XCreateWindow (xdisplay,
+                       RootWindow (xdisplay, 0),
+                       -100, -100, 1, 1,
+                       0,
+                       CopyFromParent,
+                       CopyFromParent,
+                       CopyFromParent,
+                       CWOverrideRedirect | CWEventMask,
+                       &attrs);
+      type_atom = sn_internal_atom_get (display, message_type);
+      type_atom_begin = sn_internal_atom_get (display, message_type_begin);
     
-    while (src != src_end)
       {
-        dest = &xevent.xclient.data.b[0];
-        dest_end = dest + 20;        
+        XEvent xevent;
+        const char *src;
+        const char *src_end;
+        char *dest;
+        char *dest_end;
+
+        xevent.xclient.type = ClientMessage;
+        xevent.xclient.message_type = type_atom_begin;
+        xevent.xclient.display = xdisplay;
+        xevent.xclient.window = xwindow;
+        xevent.xclient.format = 8;
+
+        src = message;
+        src_end = message + strlen (message) + 1; /* +1 to include nul byte */
         
-        while (dest != dest_end &&
-               src != src_end)
+        while (src != src_end)
           {
-            *dest = *src;
-            ++dest;
-            ++src;
+            dest = &xevent.xclient.data.b[0];
+            dest_end = dest + 20;
+
+            while (dest != dest_end &&
+                   src != src_end)
+              {
+                *dest = *src;
+                ++dest;
+                ++src;
+              }
+
+            XSendEvent (xdisplay,
+                        RootWindow (xdisplay, screen),
+                        False,
+                        PropertyChangeMask,
+                        &xevent);
+
+            xevent.xclient.message_type = type_atom;
           }
-        
-        XSendEvent (xdisplay,
-                    RootWindow (xdisplay, screen),
-                    False,
-                    PropertyChangeMask,
-                    &xevent);
-
-        xevent.xclient.message_type = type_atom;
       }
-  }
 
-  XDestroyWindow (xdisplay, xwindow);
-  XFlush (xdisplay);
+      XDestroyWindow (xdisplay, xwindow);
+      XFlush (xdisplay);
+    }
+    break;
+   case SN_DISPLAY_TYPE_XCB:
+    {
+      xcb_atom_t type_atom;
+      xcb_atom_t type_atom_begin;
+      xcb_window_t xwindow;
+      uint32_t attrs[] = { 1, XCB_EVENT_MASK_PROPERTY_CHANGE | XCB_EVENT_MASK_STRUCTURE_NOTIFY };
+      xcb_connection_t *xconnection = sn_display_get_x_connection (display);
+      xcb_screen_t *s = sn_internal_display_get_xcb_screen (display, screen);
+
+      xwindow = xcb_generate_id(xconnection);
+      xcb_create_window(xconnection, s->root_depth, xwindow, s->root,
+                        -100, -100, 1, 1, 0, XCB_COPY_FROM_PARENT, s->root_visual,
+                        XCB_CW_OVERRIDE_REDIRECT | XCB_CW_EVENT_MASK,
+                        attrs);
+
+      type_atom = sn_internal_atom_get (display, message_type);
+      type_atom_begin = sn_internal_atom_get (display, message_type_begin);
+
+      {
+        xcb_client_message_event_t xevent;
+        const char *src;
+        const char *src_end;
+        unsigned char *dest;
+        unsigned char *dest_end;
+
+        xevent.response_type = XCB_CLIENT_MESSAGE;
+        xevent.type = type_atom_begin;
+        xevent.window = xwindow;
+        xevent.format = 8;
+
+        src = message;
+        src_end = message + strlen (message) + 1; /* +1 to include nul byte */
+
+        while (src != src_end)
+          {
+            dest = &xevent.data.data8[0];
+            dest_end = dest + 20;
+
+            while (dest != dest_end &&
+                   src != src_end)
+              {
+                *dest = *src;
+                ++dest;
+                ++src;
+              }
+
+            xcb_send_event (xconnection, 0, s->root, XCB_EVENT_MASK_PROPERTY_CHANGE,
+                            (char *) &xevent);
+
+            xevent.type = type_atom;
+          }
+      }
+
+      xcb_destroy_window (xconnection, xwindow);
+      xcb_flush(xconnection);
+    }
+    break;
+  }
 }
 
 typedef struct
 {
-  Display *xdisplay;
+  void *xid;
   Atom atom;
   Window xwindow;
   sn_bool_t found_handler;
@@ -239,7 +302,7 @@ handler_for_atom_foreach (void *value,
   SnXmessageHandler *handler = value;
   HandlerForAtomData *hfad = data;
   
-  if (handler->xdisplay == hfad->xdisplay &&
+  if (handler->xid == hfad->xid &&
       (handler->type_atom == hfad->atom ||
        handler->type_atom_begin == hfad->atom))
     {
@@ -252,7 +315,8 @@ handler_for_atom_foreach (void *value,
 
 static sn_bool_t
 some_handler_handles_event (SnDisplay *display,
-                            XEvent    *xevent)
+                            Atom atom,
+                            Window win)
 {
   HandlerForAtomData hfad;
   SnList *xmessage_funcs;
@@ -260,9 +324,9 @@ some_handler_handles_event (SnDisplay *display,
   sn_internal_display_get_xmessage_data (display, &xmessage_funcs,
                                          NULL);
   
-  hfad.atom = xevent->xclient.message_type;
-  hfad.xdisplay = sn_display_get_x_display (display);
-  hfad.xwindow = xevent->xclient.window;
+  hfad.atom = atom;
+  hfad.xid = sn_internal_display_get_id (display);
+  hfad.xwindow = win;
   hfad.found_handler = FALSE;
 
   if (xmessage_funcs)
@@ -275,8 +339,7 @@ some_handler_handles_event (SnDisplay *display,
 
 typedef struct
 {
-  Display *xdisplay;
-  XEvent *xevent;
+  Window window;
   SnXmessage *message;
 } FindMessageData;
 
@@ -287,7 +350,7 @@ find_message_foreach (void *value,
   SnXmessage *message = value;
   FindMessageData *fmd = data;
   
-  if (fmd->xevent->xclient.window == message->xwindow)
+  if (fmd->window == message->xwindow)
     {
       fmd->message = message;
       return FALSE;
@@ -297,28 +360,59 @@ find_message_foreach (void *value,
 }
 
 static SnXmessage*
-add_event_to_messages (SnDisplay *display,
-                       XEvent    *xevent)
+message_new(Atom type_atom_begin, Window win)
+{
+  SnXmessage *message = sn_new0 (SnXmessage, 1);
+  message->type_atom_begin = type_atom_begin;
+  message->xwindow = win;
+  message->message = NULL;
+  message->allocated = 0;
+  return message;
+}
+
+static sn_bool_t
+message_set_message(SnXmessage *message, const char *src)
+{
+  const char *src_end;
+  char *dest;
+  sn_bool_t completed = FALSE;
+
+  src_end = src + 20;
+
+  message->message = sn_realloc (message->message,
+                                 message->allocated + (src_end - src));
+  dest = message->message + message->allocated;
+  message->allocated += (src_end - src);
+
+  /* Copy bytes, be sure we get nul byte also */
+  while (src != src_end)
+    {
+      *dest = *src;
+
+      if (*src == '\0')
+        {
+          completed = TRUE;
+          break;
+        }
+
+      ++dest;
+      ++src;
+    }
+
+  return completed;
+}
+
+static SnXmessage*
+get_or_add_message(SnList *pending_messages,
+                   Window win,
+                   Atom type_atom_begin)
 {
   FindMessageData fmd;
   SnXmessage *message;
-  const char *src;
-  const char *src_end;
-  char *dest;
-  sn_bool_t completed;
-  SnList *pending_messages;
-  
-  /* We don't want screwy situations to end up causing us to allocate
-   * infinite memory. Cap the length of a message.
-   */
-#define MAX_MESSAGE_LENGTH 4096
-  
-  fmd.xdisplay = sn_display_get_x_display (display);
-  fmd.xevent = xevent;
+
+  fmd.window = win;
   fmd.message = NULL;
 
-  sn_internal_display_get_xmessage_data (display, NULL,
-                                         &pending_messages);
   
   if (pending_messages)
     sn_list_foreach (pending_messages, find_message_foreach, &fmd);
@@ -327,17 +421,34 @@ add_event_to_messages (SnDisplay *display,
 
   if (message == NULL)
     {
-      
-      message = sn_new0 (SnXmessage, 1);
-
-      message->type_atom_begin = xevent->xclient.message_type;
-      message->xwindow = xevent->xclient.window;
-      message->message = NULL;
-      message->allocated = 0;
+      message = message_new(type_atom_begin, win);
 
       sn_list_prepend (pending_messages, message);
     }
   
+  return message;
+}
+
+static SnXmessage*
+add_event_to_messages (SnDisplay *display,
+                       Window win,
+                       Atom message_type,
+                       const char *data)
+{
+  SnXmessage *message;
+  SnList *pending_messages;
+
+  sn_internal_display_get_xmessage_data (display, NULL,
+                                         &pending_messages);
+
+  message = get_or_add_message(pending_messages,
+                               win, message_type);
+
+  /* We don't want screwy situations to end up causing us to allocate
+   * infinite memory. Cap the length of a message.
+   */
+#define MAX_MESSAGE_LENGTH 4096
+
   if (message->allocated > MAX_MESSAGE_LENGTH)
     {
       /* This message is some kind of crap - just dump it. */
@@ -347,32 +458,7 @@ add_event_to_messages (SnDisplay *display,
       return NULL;
     }
   
-  src = &xevent->xclient.data.b[0];
-  src_end = src + 20;
-  
-  message->message = sn_realloc (message->message,
-                                 message->allocated + (src_end - src));
-  dest = message->message + message->allocated;
-  message->allocated += (src_end - src);
-
-  completed = FALSE;
-  
-  /* Copy bytes, be sure we get nul byte also */
-  while (src != src_end)
-    {
-      *dest = *src;
-      
-      if (*src == '\0')
-        {
-          completed = TRUE;
-          break;
-        }
-      
-      ++dest;
-      ++src;    
-    }
-  
-  if (completed)
+  if (message_set_message (message, data))
     {
       /* Pull message out of the pending queue and return it */
       sn_list_remove (pending_messages, message);
@@ -396,7 +482,7 @@ dispatch_message_foreach (void *value,
   MessageDispatchData *mdd = data;  
   
   if (handler->type_atom_begin == mdd->message->type_atom_begin &&
-      sn_display_get_x_display (mdd->display) == handler->xdisplay)
+      sn_internal_display_get_id (mdd->display) == handler->xid)
     (* handler->func) (mdd->display,
                        handler->message_type,
                        mdd->message->message,
@@ -405,29 +491,10 @@ dispatch_message_foreach (void *value,
   return TRUE;
 }
 
-sn_bool_t
-sn_internal_xmessage_process_event (SnDisplay *display,
-                                    XEvent    *xevent)
+static void
+xmessage_process_message (SnDisplay *display, SnXmessage *message)
 {
-  sn_bool_t retval;
-  SnXmessage *message;
-
-  retval = FALSE;
-  message = NULL;
-  
-  switch (xevent->xany.type)
-    {
-    case ClientMessage:
-      if (some_handler_handles_event (display, xevent))
-        {
-          retval = TRUE;
-          
-          message = add_event_to_messages (display, xevent);
-        }
-      break;
-    }
-
-  if (message != NULL)
+  if (message)
     {
       /* We need to dispatch and free this message; ignore
        * messages containing invalid UTF-8
@@ -462,7 +529,64 @@ sn_internal_xmessage_process_event (SnDisplay *display,
       sn_free (message->message);
       sn_free (message);
     }
-  
+}
+
+sn_bool_t
+sn_internal_xmessage_process_event (SnDisplay *display,
+                                    XEvent    *xevent)
+{
+  sn_bool_t retval = FALSE;
+  SnXmessage *message = NULL;
+
+  switch (xevent->xany.type)
+    {
+    case ClientMessage:
+      if (some_handler_handles_event (display,
+                                      xevent->xclient.message_type,
+                                      xevent->xclient.window))
+        {
+          retval = TRUE;
+
+          message = add_event_to_messages (display,
+                                           xevent->xclient.window,
+                                           xevent->xclient.message_type,
+                                           xevent->xclient.data.b);
+        }
+      break;
+    }
+
+  xmessage_process_message (display, message);
+
+  return retval;
+}
+
+sn_bool_t
+sn_xcb_internal_xmessage_process_event (SnDisplay           *display,
+                                        xcb_generic_event_t *xevent)
+{
+  sn_bool_t retval = FALSE;
+  SnXmessage *message = NULL;
+  xcb_client_message_event_t *ev = (xcb_client_message_event_t *) xevent;
+
+  switch (XCB_EVENT_RESPONSE_TYPE(xevent))
+    {
+    case XCB_CLIENT_MESSAGE:
+      if (some_handler_handles_event (display,
+                                      ev->type,
+                                      ev->window))
+        {
+          retval = TRUE;
+
+          message = add_event_to_messages (display,
+                                           ev->window,
+                                           ev->type,
+                                           (const char *) ev->data.data8);
+        }
+      break;
+    }
+
+  xmessage_process_message (display, message);
+
   return retval;
 }
 
